import math
import os
import pprint
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Union

import pkg_resources
import pycrfsuite
import regex as re
from loguru import logger

from . import config, regexes
from .train import _PUNCTUATIONS, _sample_to_features
from .utils import split_keep_multiple_separators


class SentSplit():

    def __init__(self, lang: str, **kwargs: Any) -> None:
        self.lang = lang
        try:
            default_config = deepcopy(getattr(config, f'{self.lang}_config'))
        except AttributeError:
            logger.critical(f"Unsupported language: {self.lang.upper()}")
            logger.info(f"Falling back to the `base_config`. Need to provide `model` path.")
            default_config = deepcopy(getattr(config, 'base_config'))

        # override config arguments
        for k, v in kwargs.items():
            if k not in default_config:
                logger.warning(f"`{k}` not in config, skipped")
                continue
            default_config[k] = v

        self.config = default_config

        # load tagger
        model_path = self.config['model']
        if not os.path.isfile(model_path):
            model_path = pkg_resources.resource_filename('sentsplit', model_path)
        self.tagger = SentSplit._load_model(model_path)

        # fill regexes
        self._fill_regexes()

        config_string = pprint.pformat(self.config, indent=2)
        logger.info(f"SentSplit for {self.lang.upper()} loaded:\n{config_string}")

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    @staticmethod
    def _load_model(model_path: str) -> pycrfsuite.Tagger:
        tagger = pycrfsuite.Tagger()
        tagger.open(model_path)
        return tagger

    def _fill_regexes(self) -> None:
        '''
        Retrieve actual regexes from `regexes.py` by their `name`s if not given
        '''
        for rgx_index, rgx in enumerate(self.config['segment_regexes']):
            if 'regex' not in rgx:
                self.config['segment_regexes'][rgx_index] = getattr(regexes, rgx['name'])
        for rgx_index, rgx in enumerate(self.config['prevent_regexes']):
            if 'regex' not in rgx:
                self.config['prevent_regexes'][rgx_index] = getattr(regexes, rgx['name'])

    def segment(self, string: Union[str, List], strip_spaces: Union[None, bool] = None) -> List[str]:
        if strip_spaces is None:
            strip_spaces = self.config['strip_spaces']
        else:
            assert isinstance(strip_spaces, bool), "`strip_spaces` must be a boolean value"

        # for string input
        if isinstance(string, str):
            result = self._segment(string, strip_spaces)
        # for list input
        else:
            assert isinstance(string, list)
            result = [self._segment(t, strip_spaces) for t in string]
        return result

    def _segment(self, original_string: str, strip_spaces: bool) -> List[str]:
        '''This method deals with a single string'''
        # initially segment by line feeds
        strings = split_keep_multiple_separators(original_string, ['\n'])

        # list of original characters per string
        chars_strings = []
        # list of character n-gram features per string
        features_strings = []
        # list of tuples(matched_spaces, start_ind, end_ind) per string
        multiple_spaces_positions_strings = []

        for string in strings:
            # keep the original characters per string
            chars_strings.append([c for c in string])
            if self.config['handle_multiple_spaces']:
                # replace multiple spaces with a single space for better segmentation by CRF model
                preprocessed_string, multiple_spaces_positions = SentSplit._substitute_multiple_spaces(string)
                multiple_spaces_positions_strings.append(multiple_spaces_positions)
                # convert the string into character n-gram features
                features_strings.append(_sample_to_features([c for c in preprocessed_string], self.config['ngram']))
            else:
                # convert the string into character n-gram features
                features_strings.append(_sample_to_features([c for c in string], self.config['ngram']))

        # tag strings
        y_tags_strings = [self.tagger.tag(f_s) for f_s in features_strings]

        if self.config['handle_multiple_spaces']:
            # adjust y_tags_strings to account for the removed multiple spaces
            y_tags_strings = SentSplit._adjust_tags_for_multiple_spaces(y_tags_strings, multiple_spaces_positions_strings)

        y_tags_strings = SentSplit._tag_segment_regexes(y_tags_strings, strings, self.config['segment_regexes'])
        y_tags_strings = SentSplit._tag_prevent_regexes(y_tags_strings, strings, self.config['prevent_regexes'],
                                                        self.config['maxcut'], self.config['prevent_word_split'])
        results = SentSplit._segment_by_char_tag(chars_strings, y_tags_strings, strip_spaces, self.config['maxcut'], self.config['mincut'])
        return results

    @staticmethod
    def _substitute_multiple_spaces(line: str) -> Tuple[str, List[Tuple[int, int]]]:
        '''
        Substitute multiple spaces with a single space and record their indices so that they can be restored later
        multiple_spaces_positions: [(start_ind, end_ind), ..]
        '''
        rgx_multiple_spaces = r'(\s{2,})'
        multiple_spaces_positions = [(m.start(0), m.end(0)) for m in re.finditer(rgx_multiple_spaces, line)]
        if len(multiple_spaces_positions) > 0:
            line = re.sub(rgx_multiple_spaces, ' ', line)
        return line, multiple_spaces_positions

    @staticmethod
    def _adjust_tags_for_multiple_spaces(y_tags_strings: List[List[str]],
                                         multiple_spaces_positions_strings: List[List[Tuple[int, int]]]) -> List[List[str]]:
        '''
        So far y_tags_strings contains labels (tags) for the multiple-space-substituted strings
        This method adds back (`len_matched_spaces` - 1) 'O' tags after the substituted single space to restore original strings
        Note that `start_char_ind` and `end_char_ind` are character-level indices of the original strings
        Example:
            Input:
                (original) string: 'It is good.   Ha'
                preprocessed_string: 'It is good. Ha'  (after _substitute_multiple_spaces)
                y_tags_string: ['O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'EOS', 'O', 'O','O']
            Returns:
                (Restores the three spaces after '~ good.'
                y_tags_string: ['O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'O', 'EOS', 'O', 'O', 'O', 'O','O']
        '''
        assert len(y_tags_strings) == len(multiple_spaces_positions_strings)
        for string_index, multiple_spaces_positions in enumerate(multiple_spaces_positions_strings):
            y_tags = y_tags_strings[string_index]
            for start_char_ind, end_char_ind in multiple_spaces_positions:
                len_matched_spaces = end_char_ind - start_char_ind
                # y_tags already contains an 'O' for one space, hence subtract 1
                o_tags_for_matched_spaces = ['O'] * (len_matched_spaces - 1)
                # insert 'O's after the index of the substituted single space
                y_tags = y_tags[:start_char_ind + 1] + o_tags_for_matched_spaces + y_tags[start_char_ind + 1:]
            y_tags_strings[string_index] = y_tags
        return y_tags_strings

    @staticmethod
    def _tag_segment_regexes(y_tags_strings: List[List[str]], strings: List[str], segment_regexes: List[Dict]) -> List[List[str]]:
        '''
        Label either the start or end indices of the matched regex patterns with 'EOS'
        @param segment_regexes: [{'regex': '<pattern>', 'at': <'end' or 'start'>}, ..]
        '''
        for string_index, string in enumerate(strings):
            assert len(string) == len(y_tags_strings[string_index])
            for rgx in segment_regexes:
                for matched_position in re.finditer(rgx['regex'], string):
                    start = matched_position.start(0)
                    end = matched_position.end(0) - 1
                    if rgx['at'] == 'start':
                        y_tags_strings[string_index][start] = 'EOS'
                    else:
                        y_tags_strings[string_index][end] = 'EOS'
        return y_tags_strings

    @staticmethod
    def _tag_prevent_regexes(y_tags_strings: List[List[str]], strings: List[str], prevent_regexes: List[Dict],
                             maxcut: int, prevent_word_split: bool) -> List[List[str]]:
        '''
        Remove 'EOS' label for characters that are matched by prevent_regexes
        and prevent these characters from being cut due to maxcut
        @param prevent_regexes: [{'regex': '<pattern>'}, ..]
        '''
        def _tag_prevent_word_split(y_tags: List[str], curr_string: str) -> List[str]:
            '''Prevent segmentation occurring in the middle of a word'''
            assert isinstance(curr_string, str)
            for i, tag in enumerate(y_tags[:-1]):
                if tag != 'EOS':
                    continue
                char = curr_string[i]
                if char not in _PUNCTUATIONS and not char.isspace() and not curr_string[i + 1].isspace():
                    y_tags[i] = 'O'
            return y_tags

        def _get_last_segmented_index(labels: List[str], pivot: int) -> int:
            '''Return the closest previously cut index to pivot'''
            if pivot < 1:
                return -1
            for index in range(pivot - 1, -1, -1):
                if labels[index] == 'EOS':
                    return index
            return -1

        for string_index, string in enumerate(strings):
            assert len(string) == len(y_tags_strings[string_index])
            y_tags_string = y_tags_strings[string_index]
            if prevent_word_split:
                y_tags_string = _tag_prevent_word_split(y_tags_string, string)
            for rgx in prevent_regexes:
                for matched_position in re.finditer(rgx['regex'], string):
                    start = matched_position.start(0)
                    end = matched_position.end(0) - 1
                    # add 'O' at matched positions
                    for pos in range(start, end + 1):
                        y_tags_string[pos] = 'O'

                    # make sure this pattern is not maxcut later by segmenting before the pattern
                    if end < maxcut:
                        # no need to check if the pattern falls within the maxcut
                        continue
                    last_segmented_index = _get_last_segmented_index(y_tags_string, start)
                    last_segmented_plus_one = last_segmented_index + 1
                    length_to_cover = end - last_segmented_plus_one + 1  # inclusive
                    num_cuts = math.floor(length_to_cover / maxcut)
                    if num_cuts < 1:
                        num_cuts = 1
                    maxcutted_length = num_cuts * maxcut - 1
                    next_maxcut_pos = last_segmented_plus_one + maxcutted_length
                    # cut before the pattern if the next_maxcut_pos falls within the pattern
                    if start <= next_maxcut_pos < end and start - 1 > 0:
                        y_tags_string[start - 1] = 'EOS'

            y_tags_strings[string_index] = y_tags_string
        return y_tags_strings

    @staticmethod
    def _segment_by_char_tag(chars_strings: List[List[str]], y_tags_strings: List[List[str]], strip_spaces: bool, maxcut: int, mincut: int) -> List[str]:
        '''
        Loop through character-level tags and segment when 'EOS' and other conditions are met
        :param chars_strings: list of lines where each line consists of characters
        :param y_tags_strings: list of lines where each line consists of tags which are either 'O' or 'EOS'
        '''
        def _check_and_add_sentence(sent: str, str_len: int, cur_ind: int, is_leftover: bool = False) -> bool:
            '''
            Check if the given `sent` can be added to `results`.
            If so, add to `results` and return `True`; otherwise, return `False`
            '''
            if len(sent) <= mincut:
                if not is_leftover:
                    return False
                # if it is a leftover, but also an empty string, then don't add
                elif len(sent) <= 0:
                    return False
            if strip_spaces:
                sent = sent.strip()
            if len(sent) <= 0:
                return False
            # if the no. of remaining characters are less than mincut, don't add
            if not is_leftover and (str_len - cur_ind) <= mincut:
                return False
            results.append(sent)
            return True

        def _segment_maxcut_string(sent: str) -> str:
            '''
            Heuristically segment a maxcut string into two, add the first half to `results`, and return the remaining half.
            A list of heuristic regexes are applied to `sent` in decreasing order of importance.
            As soon as a matching is found, the sentence is segmented.
            '''
            rgx_punctuation_and_closing = r'(?<=[\.。︀?？!！…])[\'"’”❜❞›»❯」』)）\]］】〟]'
            rgx_closing_and_punctuation = r'(?<=[\'"’”❜❞›»❯」』)）\]］】〟])[\.。︀?？!！…]'
            rgx_space_and_dash = r'(?<=\s)\p{Pd}'
            rgx_colon = r'[:﹕：;﹔；؛⁏]'
            rgx_not_space_and_closing_and_space = r'(?<=[^\s])[\'"’”❜❞›»❯」』)）\]］】〟](?=\s)'
            rgx_comma = r'[,，﹐]'
            rgx_closing = r'[’”❜❞›»❯」』)）\]］】〟]'
            rgx_whitespace = r'\s'
            heuristics = [rgx_punctuation_and_closing, rgx_closing_and_punctuation, rgx_space_and_dash,
                          rgx_colon, rgx_not_space_and_closing_and_space, rgx_comma, rgx_closing, rgx_whitespace]
            for heu in heuristics:
                for match in re.finditer(heu, sent):
                    assert match.end() - match.start() == 1
                    first_half = sent[:match.end()]
                    if _check_and_add_sentence(first_half, len(sent), len(first_half)):
                        return sent[match.end():]
            if _check_and_add_sentence(sent, is_leftover=True):
                return ''
            raise RuntimeError(f"Cannot segment maxcut string: {sent}")

        results = []
        for char_string, tags in zip(chars_strings, y_tags_strings):
            sentence = ''
            string_length = len(char_string)
            for current_index, (current_character, tag) in enumerate(zip(char_string, tags)):
                if len(sentence) >= maxcut:
                    sentence = _segment_maxcut_string(sentence)
                if tag == 'EOS':
                    sentence += current_character
                    if _check_and_add_sentence(sentence, string_length, current_index):
                        sentence = ''
                else:
                    sentence += current_character
            if _check_and_add_sentence(sentence, string_length, current_index, is_leftover=True):
                sentence = ''
        return results

    def close(self):
        if self.tagger is not None:
            return self.tagger.close()



import random
from collections import Counter
from typing import List, Set, Tuple

import pycrfsuite
from loguru import logger
from tqdm import tqdm

from .utils import read_lines

_PUNCTUATIONS = {'.', '?', '!', '\"', '\'', '”', '．', '？', '！', '。', '…'}


def train_crf_model(corpus_path: str, ngram: int, output_path: str, sample_min_length: int, crf_max_iteration: int,
                    depunctuation_ratio: float, num_depunctuation_endings: int, ending_length: int, despace_ratio: float) -> None:
    train_samples = _preprocess(corpus_path, sample_min_length, depunctuation_ratio, num_depunctuation_endings, ending_length, despace_ratio)
    X_train, y_train = _create_features(train_samples, ngram)
    _fit_model(X_train, y_train, output_path, crf_max_iteration)


def _compute_top_k_depunctuated_endings(lines: List[str], num_depunctuation_endings: int, ending_length: int) -> Set[str]:
    """
    Computes the most common endings of sentences.
    Applicable only to languages that utilize specific endings
    that can indicate the end of sentence even without the punctuations.
    """
    assert ending_length > 0 and num_depunctuation_endings > 0
    all_endings = []
    for line in lines:
        line = line.strip()
        if len(line) - 1 < ending_length or line[-1] not in _PUNCTUATIONS:
            continue
        ending = line[-ending_length - 1:-1]
        all_endings.append(ending)
    ending_counter = Counter(all_endings)
    top_endings = ending_counter.most_common(num_depunctuation_endings)
    logger.info(f"Top-{num_depunctuation_endings} endings are: {top_endings}")
    return set(ending for ending, _ in top_endings)


def _preprocess(corpus_path: str, sample_min_length: int, depunctuation_ratio:float,
                num_depunctuation_endings: int, ending_length: int, despace_ratio: float) -> List[List[Tuple[str, str]]]:
    logger.info("Preprocessing sentences..")
    assert 0.0 <= depunctuation_ratio <= 1.0
    assert 0.0 <= despace_ratio <= 1.0

    raw_lines = read_lines(corpus_path)
    random.shuffle(raw_lines)

    num_depunctuation_remainings = int(len(raw_lines) * depunctuation_ratio)
    if depunctuation_ratio > 0.0:
        depunctuated_endings = _compute_top_k_depunctuated_endings(raw_lines, num_depunctuation_endings, ending_length)

    num_despace_remainings = int(len(raw_lines) * despace_ratio)

    train_samples = []
    single_sample = []
    for line in tqdm(raw_lines):
        chars = [c for c in line.strip()]
        labels = ['O' for _ in chars[:-1]] + ['EOS']
        # Ex. 'Hello!' -> [('H', 'O'), ('e', 'O'), ('l', 'O'), ('l', 'O'), ('o', 'O'), ('!', 'EOS')]
        char_and_labels = [(char, label) for char, label in zip(chars, labels)]

        if len(single_sample) < 1:
            single_sample = char_and_labels
        else:
            if num_depunctuation_remainings > 0 and single_sample[-1][0] in _PUNCTUATIONS and \
               len(single_sample) - 1 > ending_length and \
               ''.join(char for char, _ in single_sample[-ending_length - 1:-1]) in depunctuated_endings:
                single_sample = single_sample[:-1]
                new_eos = (single_sample[-1][0], 'EOS')
                single_sample[-1] = new_eos
                num_depunctuation_remainings -= 1
            if num_despace_remainings > 0 and single_sample[-1][0] in _PUNCTUATIONS:
                single_sample += char_and_labels
                num_despace_remainings -= 1
            else:
                single_sample += [(' ', 'O')] + char_and_labels

        if len(single_sample) >= sample_min_length:
            train_samples.append(single_sample)
            single_sample = []
    return train_samples


def _create_features(samples: List[List[Tuple[str, str]]], ngram: int) -> Tuple[List[List[List[str]]], List[List[str]]]:
    logger.info("Creating features..")
    X = []
    y = []
    for sample in tqdm(samples):
        X.append(_sample_to_features(sample, ngram))
        y.append(_sample_to_labels(sample))
    return X, y


def _sample_to_features(sample: List[Tuple[str, str]], ngram: int) -> List[List[str]]:
    def _char_to_features(i: int) -> List[str]:
        char = sample[i][0]
        feats = [
            'bias',
            f'char={char}',
            f'char.isdigit={char.isdigit()}',
            f'char.isupper={char.isupper()}',
        ]
        for j in range(1, min(i, ngram) + 1):
            char_j = sample[i - j][0]
            feats.append(f'-{j}:char={char_j}')
        for j in range(1, min(len(sample) - 1 - i, ngram) + 1):
            char_j = sample[i + j][0]
            feats.append(f'+{j}:char={char_j}')
        return feats

    features = []
    for i in range(len(sample)):
        features.append(_char_to_features(i))
    return features


def _sample_to_labels(sample: List[Tuple[str, str]]) -> List[str]:
    return [label for _, label in sample]


def _fit_model(X_train: List[List[List[str]]], y_train: List[List[str]], output_path: str, crf_max_iteration: int) -> None:
    logger.info("Fitting CRF model..")
    trainer = pycrfsuite.Trainer(verbose=True)
    for xseq, yseq in zip(X_train, y_train):
        trainer.append(xseq, yseq)
    trainer.set_params({
        'c1': 1.0,   # coefficient for L1 penalty
        'c2': 1e-3,  # coefficient for L2 penalty
        'epsilon': 1e-4,
        'max_iterations': crf_max_iteration,  # stop earlier
        # include transitions that are possible, but not observed
        'feature.possible_transitions': True
    })
    trainer.train(output_path)
    logger.info(f"Done! Model saved at {output_path}")



import os
import pprint
import subprocess
import sys
from argparse import Namespace
from copy import deepcopy
from datetime import datetime
from multiprocessing import Pool
from typing import List

from loguru import logger
from tqdm import tqdm

from sentsplit import config
from sentsplit.segment import SentSplit
from sentsplit.train import train_crf_model
from sentsplit.utils import write_lines


def sentsplit_train(args: Namespace) -> None:
    lang = args.lang
    corpus_path = args.corpus
    ngram = args.ngram
    output_path = args.output
    sample_min_length = args.sample_min_length
    crf_max_iteration = args.crf_max_iteration
    depunctuation_ratio = args.depunctuation_ratio
    num_depunctuation_endings = args.num_depunctuation_endings
    ending_length = args.ending_length
    despace_ratio = args.despace_ratio

    args_string = pprint.pformat(vars(args), indent=2)
    logger.info(f"Training a new CRF model:\n{args_string}")

    if output_path is None:
        corpus_basename = os.path.basename(corpus_path)
        today_str = datetime.today().strftime('%d%m%Y')
        output_path = f'./{corpus_basename}.{lang}-{ngram}-gram-{today_str}.model'

    train_crf_model(corpus_path, ngram, output_path, sample_min_length, crf_max_iteration, depunctuation_ratio, num_depunctuation_endings, ending_length, despace_ratio)


def call_sentsplit_batch(line: str) -> List[str]:
    return sentsplit.segment(line.rstrip('\n'))


def sentsplit_segment(args: Namespace) -> None:
    lang = args.lang
    input_file = args.input
    output_file = args.output
    assert os.path.isfile(input_file)
    if output_file is None:
        output_file = f'{input_file}.segment'

    override_options = vars(args)
    cores = override_options['cores']
    del override_options['cores']
    try:
        default_config = deepcopy(getattr(config, '{}_config'.format(lang)))
    except AttributeError:
        logger.critical(f"Unsupported language: {lang.upper()}")
        sys.exit(1)

    for k, v in default_config.items():
        if override_options[k] is not None:
            default_config[k] = override_options[k]

    global sentsplit
    sentsplit = SentSplit(lang, **default_config)

    num_lines = int(subprocess.check_output(['wc', '-l', input_file]).decode('utf8').split()[0])
    with open(input_file, 'r') as inf, open(output_file, 'w') as outf:
        cnt = 0
        if cores <= 1:
            for line in tqdm(inf, total=num_lines):
                line = line.rstrip('\n')
                segments = sentsplit.segment(line)
                for segment in segments:
                    outf.write(f"{segment}\n")
                cnt += len(segments)
        else:
            with Pool(processes=cores) as p:
                pooled_results = list(tqdm(p.imap(call_sentsplit_batch, inf, chunksize=700), total=num_lines))
            sentences = [item for sublist in pooled_results for item in sublist]
            cnt = len(sentences)
            write_lines(sentences, output_file)

        logger.info(f"{num_lines} lines are segmented into {cnt} sentences, and saved at {output_file}")
    sentsplit.close()



"""
`segment_regexes`: make sure a string is segmented at either the start or end of the matching group(s)
"""
newline = {
    'name': 'newline',
    'regex': r'\n',
    'at': 'end'
}

ellipsis = {
    'name': 'ellipsis',
    'regex': r'…(?![\!\?\.．？！])',
    'at': 'end'
}

after_semicolon = {
    'name': 'after_semicolon',
    'regex': r' *;',
    'at': 'end'
}


"""
`prevent_regexes`: make sure a string is not segmented at characters that fall within the matching group(s)
"""
liberal_url = {
    # ref. https://gist.github.com/gruber/249502#gistcomment-1328838
    'name': 'liberal_url',
    'regex': r'\b((?:[a-z][\w\-]+:(?:\/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}\/)(?:[^\s()<>]|\((?:[^\s()<>]|(?:\([^\s()<>]+\)))*\))+(?:\((?:[^\s()<>]|(?:\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?«»“”‘’]))'
}

period_followed_by_lowercase = {
    'name': 'period_followed_by_lowercase',
    'regex': r'\.(?= *[a-z])'
}
